from mpi4py import MPI
import math
import numpy as np
import time
import sys
from common import *
from rook_map import *
from rook_map_new_scheme import *

comm = MPI.COMM_WORLD
worldSize = comm.Get_size()
rank = comm.Get_rank()
processorName = MPI.Get_processor_name()
TaskMaster = 0

c = 35.0
n = int(sys.argv[1]) # the number of multiplicated matrix
size = [int(sys.argv[2]), int(sys.argv[3]), int(sys.argv[4])]
M = [i for i in range(1, n + 1)]
N, P, Q = createNPQ(n)

# workers' job
if rank != TaskMaster:
    Ap, Bp = comm.recv(source=TaskMaster, tag=100)
    ABp = np.dot(Ap, Bp)
    request = comm.send(ABp, dest=TaskMaster, tag=0)
    exit()


# master job
# no comm.barrier() because encoding on master
if rank == TaskMaster:
    threshold = P[-1] + Q[-1] + 1
    indexAB = []
    Data = []
    ABp = []
    for i in range(worldSize - 1):
        ABp.insert(i, "")

    A = []
    B = []
    rand_A = np.random.rand(size[0], size[1])
    rand_B = np.random.rand(size[1], size[2])
    for i in range(n):
        A.append(np.copy(rand_A))
    for i in range(n):
        B.append(np.copy(rand_B))
        
    ## Encoding from master
    start = time.time()
    for worker_num  in range(1, worldSize):
        Ap = honerPolinomial(A, M, P, worker_num, worldSize)
        Bp = honerPolinomial(B, N, Q, worker_num, worldSize)
        comm.send([Ap, Bp], dest=worker_num , tag=100)
    print_time(start)

    # receiving data
    start = time.time()
    while(len(indexAB) < threshold): # waiting from enough results
        st = MPI.Status()
        comm.Probe(tag=0, status = st) # detect if worker sends info
        i = st.source # take the index of worker to confirm results that come from which worker
        data = comm.recv(source = i, tag= 0)
        ABp.insert(i - 1, data) # place each worker's data in a correct order
        del ABp[i] # delete placeholder
        indexAB.append(i) # record the index of worker
    indexAB.sort() # asecnding order
    print_time(start)


    # decoding time
    start = time.time()
    AB = Numpy_decode(ABp, indexAB, [size[0], size[1]], 1, 1, 1, threshold, worldSize) #1,1, 1 is m, p, n
    end = time.time()
    print_time(start)
    
