from dataclasses import dataclass
from typing import List


@dataclass
class SubAttributes:
    sub_degree: int
    sub_optimal_entry: bool = False


@dataclass
class Attributes:
    degree: int
    sub_array: List[List[SubAttributes]]
    optimal_entry: bool = False


def recursive_function(start: int, n: int, arr: List[List[Attributes]]):
    if n > 2:
        arr[start + n - 1][n - 1].optimal_entry = True
        arr[start][n - 2].optimal_entry = True
        recursive_function(start + 1, n - 2, arr)
    else:
        if n == 2:
            arr[start][0].optimal_entry = True
            arr[start + 1][1].optimal_entry = True
            return arr
        elif n == 1:
            arr[start][0].optimal_entry = True
            return arr
        else:
            # no array
            return -1


def assign_degree(arr: List[List[Attributes]], n):
    N_values = [x for x in range(2, n+2)]
    N_values[-2], N_values[-1] = 1, n
    initial_value = 0
    for M in range(n):
        arr[M][0].degree = initial_value
        initial_value += 1
        for i in range(2, n+1):
            arr[M][i - 1].degree = arr[M][i - 2].degree + max(N_values[i-2],n-N_values[i-1]+1)


def create_Q(N, n):
    Q = [0 for i in range(0, n)]
    for i in range(1, len(N)):
        Q[i] = Q[i - 1] + max(N[i - 1], n - N[i] + 1)
    return Q


def highlight_entries(n):
    entry_list = [[Attributes(0) for j in range(n)] for x in range(n)]
    recursive_function(0, n, entry_list)
    assign_degree(entry_list, n)
    return entry_list


# For the new scheeme:
def create_N_P_Q(p, q, recovery):
    Np = [i for i in range(2, p+2)]
    Np[-2], Np[-1] = 1, p
    Nq = [i for i in range(2, q+2)]
    Nq[-2], Nq[-1] = 1, q
    Pp = [i for i in range(0, p)]
    Qp = create_Q(Np, p)
    Pq = [i for i in range(0, q)]
    Qq = create_Q(Nq, q)
    new_N = []
    new_Q = []
    new_P = []
    print(Np)
    for x in Np:
        print("Special q: ", q*x+1)
        new_N.extend(range((x-1)*q+1, q*x+1))
    for val in Qp:
        new_Q.extend([j+(val*recovery) for j in Qq])
    for val in Pq:
        new_P.extend([j+(val*recovery) for j in Pp])
    return new_N, new_P, new_Q


def assign_degree_new_scheme(arr: List[List[SubAttributes]], n, main_degree, recovery):
    N_values = [x for x in range(2, n + 2)]
    N_values[-2], N_values[-1] = 1, n
    initial_value = 0
    for M in range(n):
        arr[M][0].sub_degree = initial_value
        initial_value += 1
        for i in range(2, n + 1):
            arr[M][i - 1].sub_degree = arr[M][i - 2].sub_degree + max(N_values[i - 2], n - N_values[i - 1] + 1)
    if recovery != -1:
        for x in range(n):
            for y in range(n):
                arr[x][y].sub_degree += main_degree * recovery
        return recovery
    recovery = arr[-1][-1].sub_degree + 1
    return recovery


def recursive_function_new_scheme(start: int, n: int, arr: List[List[SubAttributes]], optimal: bool, list_of_optimal: List):
    if n > 2:
        arr[start + n - 1][n - 1].sub_optimal_entry = True
        arr[start][n - 2].sub_optimal_entry = True
        if optimal:
            list_of_optimal.extend([arr[start + n - 1][n - 1].sub_degree, arr[start][n - 2].sub_degree])
        recursive_function_new_scheme(start + 1, n - 2, arr, optimal, list_of_optimal)
    else:
        if n == 2:
            arr[start][0].sub_optimal_entry = True
            arr[start + 1][1].sub_optimal_entry = True
            if optimal:
                list_of_optimal.extend([arr[start][0].sub_degree, arr[start + 1][1].sub_degree])
            return arr
        elif n == 1:
            arr[start][0].sub_optimal_entry = True
            if optimal:
                list_of_optimal.extend([arr[start][0].sub_degree])
            return arr
        else:
            # no array
            return -1


def new_scheeme(p, q):
    n = p*q
    entry_list = [[Attributes(0, [[SubAttributes(0) for j in range(q)]
                  for i in range(q)])
                  for y in range(p)]
                  for x in range(p)]
    optimal_degrees = []
    recursive_function(0, p, entry_list)
    assign_degree(entry_list, p)
    recov = -1
    for y in range(q):
        for x in range(q):
            recov = assign_degree_new_scheme(entry_list[y][x].sub_array, p,
                                             entry_list[y][x].degree, recov)
            recursive_function_new_scheme(0, p, entry_list[y][x].sub_array,
                                          entry_list[y][x].optimal_entry,
                                          optimal_degrees)
    return entry_list, sorted(optimal_degrees), recov


'''
my_arr = highlight_entries(5)
print(my_arr[5 - 1][5 - 1].degree)
print(my_arr[0][0].optimal_entry)
for row in my_arr:
    for x in row:
        print(str(x.degree) + ", ", end="")
    print()'''
